@@ -9,6 +9,7 @@
 #include "later.h"
 #include "callback_registry.h"
 #include "timer_posix.h"
+#include "debug.h"
 
 using namespace Rcpp;
 
@@ -78,11 +79,8 @@ public:
 };
 
 static void async_input_handler(void *data) {
-  // The BEGIN_RCPP and VOID_END_RCPP macros are needed so that, if an exception
-  // occurs in any of the callbacks, destructors will still execute.
-  // https://github.com/r-lib/later/issues/12
-  // https://github.com/RcppCore/Rcpp/issues/753
-  BEGIN_RCPP
+  ASSERT_MAIN_THREAD()
+
   if (!at_top_level()) {
     // It's not safe to run arbitrary callbacks when other R code
     // is already running. Wait until we're back at the top level.
@@ -105,9 +103,31 @@ static void async_input_handler(void *data) {
   // seemed to cause R_SocketWait to hang (encountered while working with the
   // future package, trying to call value(future) with plan(multisession)).
   SuspendFDReadiness sfdr_scope;
-  
-  execCallbacks();
-  VOID_END_RCPP
+
+  // This try-catch is meant to be similar to the BEGIN_RCPP and VOID_END_RCPP
+  // macros. They are needed for two reasons: first, if an exception occurs in
+  // any of the callbacks, destructors will still execute; and second, if an
+  // exception (including R-level error) occurs in a callback and it reaches
+  // the top level in an R input handler, R appears to be unable to handle it
+  // properly.
+  // https://github.com/r-lib/later/issues/12
+  // https://github.com/RcppCore/Rcpp/issues/753
+  // https://github.com/r-lib/later/issues/31
+  try {
+    execCallbacksForTopLevel();
+  }
+  catch(Rcpp::internal::InterruptedException &e) {
+    REprintf("later: interrupt occurred while executing callback.");
+  }
+  catch(std::exception& e){
+    std::string msg = "later: exception occurred while executing callback: \n";
+    msg += e.what();
+    msg += "\n";
+    REprintf(msg.c_str());
+  }
+  catch( ... ){
+    REprintf("later: c++ exception (unknown reason) occurred while executing callback.");
+  }
 }
 
 InputHandler* inputHandlerHandle;
@@ -117,6 +137,7 @@ InputHandler* dummyInputHandlerHandle;
 // itself. The real input handler cannot remove both; otherwise a segfault
 // could occur.
 static void remove_dummy_handler(void *data) {
+  ASSERT_MAIN_THREAD()
   removeInputHandler(&R_InputHandlers, dummyInputHandlerHandle);
   close(dummy_pipe_in);
   close(dummy_pipe_out);
@@ -124,6 +145,7 @@ static void remove_dummy_handler(void *data) {
 
 void ensureInitialized() {
   if (!initialized) {
+    REGISTER_MAIN_THREAD()
     buf = malloc(BUF_SIZE);
     
     int pipes[2];
@@ -154,6 +176,7 @@ void ensureInitialized() {
 }
 
 void deInitialize() {
+  ASSERT_MAIN_THREAD()
   if (initialized) {
     removeInputHandler(&R_InputHandlers, inputHandlerHandle);
     close(pipe_in);
@@ -166,6 +189,7 @@ void deInitialize() {
 }
 
 void doExecLater(Rcpp::Function callback, double delaySecs) {
+  ASSERT_MAIN_THREAD()
   callbackRegistry.add(callback, delaySecs);
   
   timer.set(*callbackRegistry.nextTimestamp());
